package polyglot.ext.polyj;

import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.frontend.Compiler; // to avoid ambiguity with java.lang.Compiler
import polyglot.main.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import polyglot.ext.polyj.ast.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.visit.*;

import java.io.*;
import java.util.*;

/**
 * This is the <code>ExtensionInfo</code> for the PolyJ language.
 *
 * Compilation Passes and visitors:
 *  1. parse
 *  2. build types (TypeBuilder)
 * ------------------barrier
 *  3. clean super (AmbiguityRemover)
 * ------------------barrier
 *  4. clean signatures (AmbiguityRemover)
 *  5. add members (AddMemberVisitor)
 * ------------------barrier
 *  6. disambiguate (AmbiguityRemover)
 *  7. constant folding (ConstantFolder)
 * ------------------barrier
 *  *. instantiation checking (PolyJInstantiationChecker)
 * ------------------barrier
 *  8. type checking (TypeChecker)
 *  9. exception checking (ExceptionChecker)
 * ------------------barrier
 *  *. base bindings generation (BaseBindingsGenerator)
 * ------------------barrier
 *  *. instantiation bindings generation (InstBindingsGenerator)
 * ------------------barrier
 * 10. serialization (ClassSerializer), optional
 * 11. translation (Translator)
 */
public class ExtensionInfo extends polyglot.ext.jl.ExtensionInfo {
    static { Topics t = new Topics(); }

    public String defaultFileExtension() {
        return "pj";
    }

    public String compilerName() {
        return "polyj";
    }

    public polyglot.main.Version version() {
        return new Version();
    }

    protected TypeSystem createTypeSystem() {
        return new PolyJTypeSystem_c();
    }

    protected NodeFactory createNodeFactory() {
        return new PolyJNodeFactory_c();
    }

    public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
        polyglot.lex.Lexer lexer;
        java_cup.runtime.lr_parser grm;

        lexer = new polyglot.ext.polyj.parse.Lexer_c(reader, source.name(), eq);
        grm = new polyglot.ext.polyj.parse.Grm(lexer, ts, nf, eq);

        return new CupParser(grm, source, eq);
    }

    public static final Pass.ID INST_CHECK = new Pass.ID("inst-check");
    public static final Pass.ID INST_CHECK_ALL = new Pass.ID("inst-check-barrier");
    public static final Pass.ID PRE_TRANS_ALL = new Pass.ID("pre-polyj-trans-all");
    public static final Pass.ID BASE_BIND = new Pass.ID("base-bind");
    public static final Pass.ID BASE_BIND_ALL = new Pass.ID("base-bind-all");
    public static final Pass.ID INST_BIND = new Pass.ID("inst-bind");
    public static final Pass.ID INST_BIND_ALL = new Pass.ID("inst-bind-all");
    public static final Pass.ID PRE_POLYJ_OUT = new Pass.ID("pre-polyj-output");
    public static final Pass.ID PRE_POLYJ_OUT_ALL = new Pass.ID("pre-polyj-output-all");

    public List passes(Job job) {
        List instCheckPasses = new ArrayList(2);
        List newExpPasses = new ArrayList(7);

        instCheckPasses.add(new VisitorPass(INST_CHECK, job,
            new InstantiationChecker(job, (PolyJTypeSystem) ts, nf)));
        instCheckPasses.add(new BarrierPass(INST_CHECK_ALL, job));

        newExpPasses.add(new BarrierPass(PRE_TRANS_ALL, job));
        newExpPasses.add(new VisitorPass(BASE_BIND, job,
            new BaseBindingsGenerator(job, (PolyJTypeSystem) ts, nf)));
        newExpPasses.add(new BarrierPass(BASE_BIND_ALL, job));
        newExpPasses.add(new VisitorPass(INST_BIND, job,
            new InstBindingsGenerator((PolyJTypeSystem) ts)));
        newExpPasses.add(new BarrierPass(INST_BIND_ALL, job));
        newExpPasses.add(new VisitorPass(PRE_POLYJ_OUT, job, 
            new PreTranslator()));
        newExpPasses.add(new BarrierPass(PRE_POLYJ_OUT_ALL, job));

        List passes = super.passes(job);
        beforePass(passes, Pass.TYPE_CHECK, instCheckPasses);
        beforePass(passes, Pass.PRE_OUTPUT_ALL, newExpPasses);
        replacePass(passes, Pass.OUTPUT,
                    new OutputPass(Pass.OUTPUT, job,
                                   new PolyJTranslator(job,
                                                       (PolyJTypeSystem) ts,
                                                       nf, targetFactory())));

        // XXX: remove these passes until Steve (?!) fixes them.
        removePass(passes, Pass.REACH_CHECK);
        removePass(passes, Pass.EXIT_CHECK);
        removePass(passes, Pass.INIT_CHECK);
        
        return passes;
    }
}
